#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <SDL/SDL.h>
#undef	main

#include "caph.h"
#include "draw.h"
#include "phys.h"

#if defined(_WIN32)
# include <windows.h>
#else
# include <unistd.h>
#endif

#define SYS_DATA_DIR	"../data/"

static void 
sys_chdir(const char *name)
{
#if defined(_WIN32)
	SetCurrentDirectory(name);
#else
	chdir(name);
#endif
}

static void
sys_sleep(long msec)
{
#if defined(_WIN32)
	Sleep(msec);
#else
	usleep(msec * 1000);
#endif
}

int		screen_w;
int		screen_h;
uint8_t		*screen_pixels;

#define	STRIP_SIZE		(512)

static point_t  *p_strip;
static int	p_it;
static int	p_state;
static int	p_flags;

static void
phys_paint_click(float x, float y, int state)
{
	convex_t	*cx;

	if (p_state && !state) {

		if (p_it > 2) {
			cx = phys_alloc(p_it + 1);
			cx->flags = p_flags;
			memcpy(cx->pts, p_strip, sizeof(point_t) * cx->pt);
			phys_insert(cx);
		}

		p_it = 0;
	}

	if (!p_state && state) {
		p_strip[p_it].x = x;
		p_strip[p_it].y = y;
	}

	p_state = state;
}

static void
phys_paint_move(float x, float y)
{
	float		dist;
	float		dx, dy;

	if (!p_state)
		return;

	dx = x - p_strip[p_it].x;
	dy = y - p_strip[p_it].y;

	dist = sqrt(dx * dx + dy * dy);

	if (dist < 8.0f)
		return;

	if (p_it > 1 && dist < 40.0f) {

		float		pdx, pdy;

		pdx = p_strip[p_it].x - p_strip[p_it - 1].x;
		pdy = p_strip[p_it].y - p_strip[p_it - 1].y;

		dist = sqrt(fabsf(pdx * dy - pdy * dx));

		if (dist < 8.0f)
			return;
	}

	if (p_it < (STRIP_SIZE - 1)) {
		p_strip[++p_it].x = x;
		p_strip[p_it].y = y;
	}
}

static void
phys_paint_draw()
{
	color_t		col = { 255, 255, 155, 0 };
	int		i;

	if (p_it > 1) {
		for (i = 1; i <= p_it; ++i) {
			draw_line(p_strip + i - 1,
					p_strip + i, col);
		}
	}
}

static void
phys_flags_draw()
{
	printf("flags: %s %s %s \n",
			(p_flags & CONVEX_CONST) ? "const" : "NULL",
			(p_flags & CONVEX_SOFT) ? "soft" : "NULL",
			(p_flags & CONVEX_BEND) ? "bend" : "NULL");
}

int main(int argc, char *argv[])
{
	SDL_Surface *screen;
	SDL_Event event;
	int fs = 0;
	int run = 1;

	sys_chdir(SYS_DATA_DIR);

	/*FILE *conf = fopen("caph.conf", "r");

	fscanf(conf, "%i %i %i",
			screen_w,
			screen_h,
			fs);

	fclose(fd);*/

	screen_w = 640;
	screen_h = 480;

	if (SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "%s:%i [PANIC] SDL_Init failed: %s \n",
				__FILE__, __LINE__, SDL_GetError());
		exit(EXIT_FAILURE);
	}

	screen = SDL_SetVideoMode(screen_w, screen_h, 32,
			(fs ? SDL_FULLSCREEN : 0) |
			SDL_DOUBLEBUF |
			SDL_HWSURFACE);

	if (screen == NULL) {
		fprintf(stderr, "%s:%i [PANIC] SDL_SetVideoMode failed: %s \n",
				__FILE__, __LINE__, SDL_GetError());
		SDL_Quit();
		exit(EXIT_FAILURE);
	}

	screen_pixels = screen->pixels;

	p_strip = malloc(sizeof(point_t) * STRIP_SIZE);
	p_it = 0;
	p_state = 0;
	p_flags = 0;

	phys_open();

	convex_t *cx = phys_alloc(5);
	cx->flags = CONVEX_CONST;
	cx->pts[0].x = 8.0f;
	cx->pts[0].y = 8.0f;
	cx->pts[1].x = 8.0f;
	cx->pts[1].y = (float) (screen_h - 8);
	cx->pts[2].x = (float) (screen_w - 8);
	cx->pts[2].y = (float) (screen_h - 8);
	cx->pts[3].x = (float) (screen_w - 8);
	cx->pts[3].y = 8.0f;
	cx->pts[4].x = 8.0f;
	cx->pts[4].y = 8.0f;
	phys_insert(cx);

	int dt, t = 0;
	int fr_n = 0;
	int fr_tval = 0;
	int fps = 0;
	int dt_min = (1000/60);
	uint32_t t0, t1;

	t0 = t1 = SDL_GetTicks();

	do {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_QUIT:
					run = 0;
					break;

				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_q)
						run = 0;
					else if (event.key.keysym.sym == SDLK_s)
						p_flags ^= CONVEX_SOFT;
					else if (event.key.keysym.sym == SDLK_c)
						p_flags ^= CONVEX_CONST;
					else if (event.key.keysym.sym == SDLK_b)
						p_flags ^= CONVEX_BEND;
					else if (event.key.keysym.sym == SDLK_d)
						phys_remove_back();
					phys_flags_draw();
					break;

				case SDL_MOUSEBUTTONDOWN:
				case SDL_MOUSEBUTTONUP:
					if (event.button.button == SDL_BUTTON_LEFT) {
						phys_paint_click((float) event.button.x,
								(float) event.button.y,
								event.type == SDL_MOUSEBUTTONDOWN);
					}
					break;

				case SDL_MOUSEMOTION:
					phys_paint_move((float) event.motion.x,
							(float) event.motion.y);
					break;
			}
		}

		t0 = SDL_GetTicks();
		dt = t0 - t1;

		while (dt < dt_min) {
			sys_sleep(1);
			t0 = SDL_GetTicks();
			dt = t0 - t1;
		}

		t1 = t0;

		++fr_n;
		t += dt;
		fr_tval += dt;

		if (fr_tval > 1000) {
			fps = (fr_n * 1000) / fr_tval;
			printf("fps: %i \n", fps);
			fflush(stdout);
			fr_n = 0;
			fr_tval = 0;
		}

		//phys_update((float)dt * 0.001f);

		if (SDL_LockSurface(screen) == 0) {

			color_t clr = {0, 0, 0, 0};

			draw_clear(clr);

			phys_update((float)dt * 0.001f);
	
			phys_draw();
			phys_paint_draw();

			SDL_UnlockSurface(screen);
			SDL_Flip(screen);
		}

	} while (run);

	phys_close();

	SDL_Quit();
}

