#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "phys.h"
#include "caph.h"

#define	PHYS_DELTA	0.0078125f

float			phys_g = 0.1f;

static convex_t		*all;
static float		adt;

static void
all_list_insert(convex_t *cx)
{
	if (all) {
		cx->next = all->next;
		all->next->prev = cx;
		cx->prev = all;
		all->next = cx;
	}
	else {
		all = cx;
		all->prev = all;
		all->next = all;
	}
}

static void
all_list_remove(convex_t *cx)
{
	if (all == all->next)
		all = NULL;
	else {
		if (all == cx)
			all = all->prev;
		cx->prev->next = cx->next;
		cx->next->prev = cx->prev;
	}
}

convex_t *phys_alloc(int len)
{
	convex_t	*cx;

	cx = malloc(sizeof(convex_t));
	cx->pt = len;
	cx->hold = NULL;
	cx->lns = NULL;

	cx->last = malloc(sizeof(point_t) * len);
	cx->pts = malloc(sizeof(point_t) * len);

	return cx;
}

void phys_free(convex_t *cx)
{
	free(cx->pts);
	free(cx->last);
	if (cx->lns) free(cx->lns);
	if (cx->hold) free(cx->hold);
	free(cx);
}

void phys_insert(convex_t *cx)
{
	float		x, y;
	float		dist;
	int		pt, ln;
	int		i, j;

	all_list_insert(cx);

	pt = cx->pt;
	ln = 0;

	for (i = 0; i < pt; ++i) {
		cx->last[i].x = cx->pts[i].x;
		cx->last[i].y = cx->pts[i].y;
	}

	if (cx->flags & CONVEX_CONST) {
		cx->hold = malloc(sizeof(link_t) * pt);

		for (i = 0; i < pt; ++i) {
			cx->hold[i].x = cx->pts[i].x;
			cx->hold[i].y = cx->pts[i].y;
		}
	}
	else if (cx->flags & CONVEX_SOFT) {
		cx->lns = malloc(sizeof(link_t) * (pt - 1));

		for (i = 1; i < pt; ++i) {
			x = cx->pts[i - 1].x - cx->pts[i].x;
			y = cx->pts[i - 1].y - cx->pts[i].y;
			dist = sqrt(x * x + y * y);

			cx->lns[ln].beg = cx->pts + i - 1;
			cx->lns[ln].end = cx->pts + i;
			cx->lns[ln++].dist = dist;
		}
	}
	else {
		int	ln_c = pt * (pt - 1) / 2;

		cx->lns = malloc(sizeof(link_t) * (ln_c));

		for (i = 0; i < pt; ++i) {

			for (j = i + 1; j < pt; ++j) {
				x = cx->pts[i].x - cx->pts[j].x;
				y = cx->pts[i].y - cx->pts[j].y;
				dist = sqrt(x * x + y * y);

				cx->lns[ln].beg = cx->pts + i;
				cx->lns[ln].end = cx->pts + j;
				cx->lns[ln++].dist = dist;
			}
		}
	}

	cx->ln = ln;
}

void phys_remove(convex_t *cx)
{
	all_list_remove(cx);
}

void phys_remove_back()
{
	all_list_remove(all->next);
}

static void
phys_verlet(convex_t *cx)
{
	point_t		*last = cx->last;
	point_t		*pts = cx->pts;
	int		i, pt = cx->pt;
	float		x, y;

	for (i = 0; i < pt; ++i) {
		x = pts[i].x;
		y = pts[i].y;
		pts[i].x = x + (x - last[i].x);
		pts[i].y = y + (y - last[i].y) + phys_g;
		last[i].x = x;
		last[i].y = y;
	}
}

static void
phys_hold(convex_t *cx)
{
	point_t		*hold = cx->hold;
	point_t		*pts = cx->pts;
	int		i, pt = cx->pt;

	for (i = 0; i < pt; ++i) {
		pts[i].x = hold[i].x;
		pts[i].y = hold[i].y;
	}
}

static void
phys_links(convex_t *cx)
{
	point_t		*beg, *end;
	link_t		*lns = cx->lns;
	int		ln = cx->ln;
	float		dx, dy;
	float		dist, diff;
	int		i, it;

	if (cx->flags & CONVEX_BEND) {
		for (it = 0; it < 8; ++it) {
			for (i = 0; i < ln; ++i) {
				beg = lns[i].beg;
				end = lns[i].end;
				dx = beg->x - end->x;
				dy = beg->y - end->y;
				dist = sqrt(dx * dx + dy * dy);
				diff = (dist - lns[i].dist);
				if (fabsf(diff) > 1.0f)	lns[i].dist += diff * 0.4f;
				diff = (lns[i].dist - dist);
				diff *= 0.25f;
				dx /= dist;
				dy /= dist;
				beg->x += dx * diff;
				beg->y += dy * diff;
				end->x += -dx * diff;
				end->y += -dy * diff;
			}
		}
	}
	else {
		for (it = 0; it < 8; ++it) {
			for (i = 0; i < ln; ++i) {
				beg = lns[i].beg;
				end = lns[i].end;
				dx = beg->x - end->x;
				dy = beg->y - end->y;
				dist = sqrt(dx * dx + dy * dy);
				diff = (lns[i].dist - dist) * 0.25f;
				dx /= dist;
				dy /= dist;
				beg->x += dx * diff;
				beg->y += dy * diff;
				end->x += -dx * diff;
				end->y += -dy * diff;
			}
		}
	}
}

#define	PHYS_SBIAS	2.0f
#define	PHYS_QSBIAS	(PHYS_SBIAS * PHYS_SBIAS)

static void
phys_push_lines(
		point_t *pt,
		const point_t *pt_p,
		point_t *la,
		point_t *lb,
		const point_t *la_p,
		const point_t *lb_p)
{
	point_t		n, s;
	float		dist, frct;
	float		d;

	n.y = (la->x - lb->x);
	n.x = -(la->y - lb->y);
	d = sqrt(n.x * n.x + n.y * n.y);
	n.x /= d;
	n.y /= d;
	d = la->x * n.x + la->y * n.y;

	dist = pt->x * n.x + pt->y * n.y - d;
	dist *= 0.701f;

	s.x = (pt->x - pt_p->x) - ((la->x - la_p->x) - (lb->x - lb_p->x)) * 0.5f;
	s.y = (pt->y - pt_p->y) - ((la->y - la_p->y) - (lb->y - lb_p->y)) * 0.5f;

	frct = s.x * n.y - s.y * n.x;
	frct *= - 0.4f;

	pt->x +=	- n.x * dist	+ n.y * frct;
	pt->y +=	- n.y * dist	- n.x * frct;
	la->x +=	+ n.x * dist	- n.y * frct;
	la->y +=	+ n.y * dist	+ n.x * frct;
	lb->x +=	+ n.x * dist	- n.y * frct;
	lb->y +=	+ n.y * dist	+ n.x * frct;
}

#define	PHYS_EPS	0.00001f
#define NONZERO(x)	(fabsf(x) > PHYS_EPS)

static int
phys_intersect_lines(
		point_t *pt,
		const point_t *pt_p,
		point_t *la,
		point_t *lb,
		const point_t *la_p,
		const point_t *lb_p)
{
	point_t		pa, ba, ppaa, bbaa;
	point_t		n_pt, n_a, n_b;
	point_t		n, pb;
	float		a, b, c, d;
	float		t, t1, t2;

	pa.x = pt_p->x - la_p->x;
	pa.y = pt_p->y - la_p->y;

	ba.y = lb_p->x - la_p->x;
	ba.x = -(lb_p->y - la_p->y);

	ppaa.x = (pt->x - pt_p->x) - (la->x - la_p->x);
	ppaa.y = (pt->y - pt_p->y) - (la->y - la_p->y);

	bbaa.y = (lb->x - lb_p->x) - (la->x - la_p->x);
	bbaa.x = -((lb->y - lb_p->y) - (la->y - la_p->y));

	a = ppaa.x * bbaa.x + ppaa.y * bbaa.y;
	b = pa.x * bbaa.x + pa.y * bbaa.y + ba.x * ppaa.x + ba.y * ppaa.y;
	c = pa.x * ba.x + pa.y * ba.y;

	if (NONZERO(a)) {
		d = b * b - 4.0f * a * c;
		if (d < 0.0f) return 0;
		d = sqrt(d);
		t1 = (- b + d) / (2.0f * a);
		t2 = (- b - d) / (2.0f * a);
		if (t1 >= 0.0f && t1 <= 1.0f) {
			if (t2 >= 0.0f && t2 <= 1.0f)
				t = (t1 < t2) ? t1 : t2;
			else
				t = t1;
		}
		else if (t2 >= 0.0f && t2 <= 1.0f)
			t = t2;
		else
			return 0;
	}
	else if (NONZERO(b)) {
		t = - c / b;
		if (t < 0.0f || t > 1.0f) return 0;	
	}
	else return 0;

	if (isnan(t)) return 0;

	n_pt.x = pt_p->x + (pt->x - pt_p->x) * t;
	n_pt.y = pt_p->y + (pt->y - pt_p->y) * t;

	n_a.x = la_p->x + (la->x - la_p->x) * t;
	n_a.y = la_p->y + (la->y - la_p->y) * t;

	n_b.x = lb_p->x + (lb->x - lb_p->x) * t;
	n_b.y = lb_p->y + (lb->y - lb_p->y) * t;

	n.x = n_b.x - n_a.x;
	n.y = n_b.y - n_a.y;

	pa.x = n_pt.x - n_a.x;
	pa.y = n_pt.y - n_a.y;
	d = pa.x * n.x + pa.y * n.y;

	pb.x = n_pt.x - n_b.x;
	pb.y = n_pt.y - n_b.y;
	d *= pb.x * n.x + pb.y * n.y;

	if (d > 0.0f) return 0;

	phys_push_lines(pt, pt_p, la, lb, la_p, lb_p);

	return 1;
}

static int it;

static void
phys_collise(const convex_t *cx, const convex_t *bound)
{
	point_t		*lst = cx->last;
	point_t		*pts = cx->pts;
	int		pt = cx->pt;
	point_t		*b_lst = bound->last;
	point_t		*b_pts = bound->pts;
	int		b_pt = bound->pt;
	int		i, j, it;

	/*color_t		red = {0, 0, 255, 0};
	color_t		blue = {255, 0, 0, 0};
	color_t		green = {0, 255, 0, 0};

	point_t		pp = { 400, 300 };
	point_t		p = { 200, 100 };

	point_t		ap = { 300, 200 };
	point_t		bp = { 100, 200 };
	point_t		a = { 300, 200 };
	point_t		b = { 100, 200 };

	draw_line(&pp, &p, green);
	draw_line(&ap, &a, blue);
	draw_line(&bp, &b, blue);
	draw_line(&ap, &bp, green);
	draw_line(&a, &b, green);

	phys_intersect_lines(
			&p,
			&pp,
			&a,
			&b,
			&ap,
			&bp);

	draw_line(&pp, &p, red);
	draw_line(&a, &b, red);

	//return ;*/

	for (i = 0; i < pt; ++i) {

		do {
			for (j = 1; j < b_pt; ++j) {
				it += phys_intersect_lines(
						pts + i,
						lst + i,
						b_pts + j - 1,
						b_pts + j,
						b_lst + j - 1,
						b_lst + j);
			}
		}
		while (0);
	}
}

static void
phys_collise_all(const convex_t *cx)
{
	convex_t	*curr = all;

	do {
		if (curr != cx) {
			phys_collise(cx, curr);
		}

		curr = curr->next;
	}
	while (curr != all);
}

void phys_open()
{
	all = NULL;
	adt = 0.0f;
}

void phys_close()
{
	//FIXME:	
}

static void
phys_update_all()
{
	convex_t	*curr = all;

	if (!curr)
		return;

	do {
		phys_verlet(curr);

		if (curr->flags & CONVEX_CONST)
			phys_hold(curr);
		else
			phys_links(curr);

		phys_collise_all(curr);

		curr = curr->next;
	}
	while (curr != all);

	/*int itlim = 8;

	do {
		it = 0;
		curr = all;

		do {
			phys_collise_all(curr);
			curr = curr->next;
		}
		while (curr != all);

		itlim += -1;
	}
	while (it && itlim);*/
}

void phys_update(float dt)
{
	adt += dt;

	if (adt > (4.0f * PHYS_DELTA))
		adt = (4.0f * PHYS_DELTA);

	while (adt > PHYS_DELTA) {
		phys_update_all();
		adt -= PHYS_DELTA;
	}
}

static void
phys_draw_convex(convex_t *cx)
{
	color_t		col_def = { 55, 55, 255, 0 };
	color_t		col_const = { 55, 255, 55, 0 };
	color_t		col_soft = { 255, 155, 155, 0 };
	color_t		col;
	point_t		*pts = cx->pts;
	int		i, pt = cx->pt;

	if (cx->flags & CONVEX_CONST)
		col = col_const;
	else if (cx->flags & CONVEX_SOFT)
		col = col_soft;
	else
		col = col_def;

	for (i = 1; i < pt; ++i) {
		draw_line(pts + i - 1,
				pts + i, col);
	}
}

void phys_draw()
{
	convex_t	*curr = all;

	if (!curr)
		return;

	do {
		phys_draw_convex(curr);
		curr = curr->next;
	}
	while (curr != all);
}

